"""
Created on Mon Sep 30 22:09:40 2019.

@author: mtageld
"""
from histomicstk.utils.general_utils import Base_HTK_Class


class Slide_iterator(Base_HTK_Class):
    """Iterate through large_image items in a girder folder."""

    def __init__(self, gc, source_folder_id, **kwargs):
        """Init Slide_iterator object.

        Arguments:
        ---------
        gc : object
            girder client object
        source_folder_id : str
            girder ID of folder in which slides are located
        keep_slides : list
            List of slide names to keep. If None, all are kept.
        discard_slides : list
            List of slide names to discard.
        kwargs : key-value pairs
            The following are already assigned defaults by Base_HTK_Class
            but can be passed here to override defaults
            [verbose, monitorPrefix, logger, logging_savepath,
            suppress_warnings]

        """
        default_attr = {
            'keep_slides': None,
            'discard_slides': [],
        }
        default_attr.update(kwargs)
        super().__init__(default_attr=default_attr)

        # set attribs
        self.gc = gc
        self.source_folder_id = source_folder_id
        self.set_slide_ids()

    def set_slide_ids(self):
        """Get dict of slide idx, indexed by name."""
        resp = self.gc.get(
            'item?folderId=%s&limit=1000000' % self.source_folder_id)
        self.slide_ids = {j['name']: j['_id'] for j in resp}
        # find discard ids
        if self.keep_slides is not None:
            discard = set(self.slide_ids.keys()) - set(self.keep_slides)
            self.discard_slides.extend(list(discard))
            self.discard_slides = list(set(self.discard_slides))
        # only keep what's relevant
        for sn in self.discard_slides:
            del self.slide_ids[sn]

    def run(self):
        """Yield information on one slide at a time."""
        for sname, sid in self.slide_ids.items():
            try:
                slide_info = self.gc.get('item/%s/tiles' % sid)
            except Exception as e:
                print(str(e))
                slide_info = {}
            slide_info['name'] = sname
            slide_info['_id'] = sid
            yield slide_info


class Workflow_runner(Base_HTK_Class):
    """Run workflow for all slides in a girder folder."""

    def __init__(
            self, slide_iterator, workflow, workflow_kwargs,
            recursive=False, catch_exceptions=True, **kwargs):
        """Init Workflow_runner object.

        Arguments:
        ---------
        slide_iterator : object
            Slide_iterator object
        workflow : method
            method whose parameters include slide_id and monitorPrefix,
            which is called for each slide
        workflow_kwargs : dict
            keyword arguments for the workflow method
        recursive : bool
            whether to run the workflow recursively on all subfolders
        catch_exceptions : bool
            whether to catch exceptions. You may want to set to false if
            for example you want to run with a debugger
        kwargs : key-value pairs
            The following are already assigned defaults by Base_HTK_Class
            but can be passed here to override defaults
            [verbose, monitorPrefix, logging_savepath, suppress_warnings]

        """
        default_attr = {}
        default_attr.update(kwargs)
        super().__init__(default_attr=default_attr)

        # set attribs
        self.workflow = workflow
        self.workflow_kwargs = workflow_kwargs
        self.recursive = recursive
        self.catch_exceptions = catch_exceptions
        if self.keep_log:
            self.exception_path = self.logname.replace(
                '.log', '_EXCEPTIONS.log')
        self.slide_iterator = slide_iterator
        self.gc = self.slide_iterator.gc
        self.si = slide_iterator.run()
        self.originalPrefix = self.monitorPrefix

    def run(self):
        """Run workflow for all slides."""
        self.n_slides = len(self.slide_iterator.slide_ids)

        def _run_slide(self, monitorStr):

            slide_info = next(self.si)
            monitorStr += ' (%s)' % (slide_info['name'])

            _ = self.workflow(
                slide_id=slide_info['_id'], monitorPrefix=monitorStr,
                **self.workflow_kwargs)

        for sno in range(self.n_slides):

            monitorStr = '%s: slide %d of %d' % (
                self.monitorPrefix, sno + 1, self.n_slides)

            if not self.catch_exceptions:
                _run_slide(self, monitorStr)
            else:
                try:
                    _run_slide(self, monitorStr)

                except Exception as e:

                    if self.keep_log:
                        self.cpr1.logger.exception(
                            '%s: SEE EXCEPTIONS FILE: %s', (
                                monitorStr, self.exception_path))
                        with open(self.exception_path, 'a') as f:
                            print(str(e))
                            f.write('%s\n' % monitorStr)
                            f.write(e.__repr__())
                            f.write('\n---------------------------------\n')
                    else:
                        print(e.__repr__())

        if self.recursive:
            # for each subfolder, call self
            for folder in self.gc.listFolder(
                    parentId=self.slide_iterator.source_folder_id):

                fpath = self.gc.get('/folder/%s/rootpath' % folder['_id'])
                fpath = '/'.join(
                    [j['object']['name'] for j in fpath],
                ) + '/' + folder['name'] + '/'

                self.monitorPrefix = '%s: %s' % (self.originalPrefix, fpath)

                # update slide iterator for subfolder
                self.slide_iterator.source_folder_id = folder['_id']
                self.slide_iterator.set_slide_ids()
                self.si = self.slide_iterator.run()

                # recurse
                self.run()


class Annotation_iterator(Base_HTK_Class):
    """Iterate through annotations in a girder item (slide)."""

    def __init__(
            self, gc, slide_id, callback=None, callback_kwargs=None, **kwargs):
        """Init Annotation_iterator object.

        Arguments:
        ---------
        gc : object
            girder client object
        slide_id : str
            girder ID of slide (item)
        callback : function
            function to apply to each annotation. Must accept at least
            the parameters "gc" and "annotation" and these will be passed
            internally to it.
        callback_kwargs : dict
            kwargs to pass to the callback (other than gc and annotation)
        kwargs : key-value pairs
            The following are already assigned defaults by Base_HTK_Class
            but can be passed here to override defaults
            [verbose, monitorPrefix, logger, logging_savepath,
            suppress_warnings]

        """
        default_attr = {}
        default_attr.update(kwargs)
        super().__init__(default_attr=default_attr)

        # set attribs
        self.gc = gc
        self.slide_id = slide_id
        self.callback = callback
        self.callback_kwargs = callback_kwargs

        # get annotations for slide
        self.slide_annotations = self.gc.get(
            '/annotation/item/' + self.slide_id)
        self.n_annotations = len(self.slide_annotations)

    def yield_callback_output_for_annotation(self):
        """Yield callback output for one annotation at a time."""
        # yield one annotation at a time
        for annidx, ann in enumerate(self.slide_annotations):
            if self.verbose > 0:
                print('%s: annotation %d of %d' % (
                    self.monitorPrefix, annidx + 1, self.n_annotations))
            if self.callback is None:
                yield ann
            else:
                yield self.callback(
                    gc=self.gc, annotation=ann, **self.callback_kwargs)

    def apply_callback_to_all_annotations(self):
        """Apply callback to all annotations and return output list."""
        runner = self.yield_callback_output_for_annotation()
        outputs = []
        for _annidx in range(self.n_annotations):
            outputs.append(next(runner))
        return outputs



import numpy as np

cimport cython
cimport numpy as np


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _max_clustering_cython(double[:, :] im not None, int[:, :] im_fgnd_mask not None, double rad):

    cdef long sx = im.shape[1]
    cdef long sy = im.shape[0]
    cdef long r = int(rad + 0.5)

    # find foreground pixels
    cdef long[:] px, py

    py, px = np.nonzero(im_fgnd_mask)

    cdef long num_pixels = py.shape[0]

    # find local maximum of all requested pixels
    cdef double[:, ::1] local_max_val = np.zeros([sy, sx], dtype=float)
    cdef long[:, ::1] local_max_ind = np.zeros([sy, sx], dtype=np.int64)
    cdef int[:, ::1] peak_found = np.zeros([sy, sx], dtype=np.int32)
    cdef double min_im_val = np.min(im)

    cdef long ox, oy, mx, my, cx, cy, nx, ny
    cdef double mval, cval, nval
    cdef int changed

    with nogil:
        for i in range(num_pixels):

            cx = px[i]

            if cx < 0 or cx >= sx:
                continue

            cy = py[i]

            if cy < 0 or cy >= sy:
                continue

            cval = im[cy, cx]

            my = cy
            mx = cx
            mval = cval

            changed = 0

            for ox in range(-r, r+1):

                nx = cx + ox

                if nx < 0 or nx >= sx:
                    continue

                for oy in range(-r, r+1):

                    if (ox * ox + oy * oy) > rad * rad:
                        continue

                    ny = cy + oy

                    if ny < 0 or ny >= sy:
                        continue

                    nval = min_im_val

                    if im_fgnd_mask[ny, nx]:
                        nval = im[ny, nx]

                    if nval > mval:

                        changed = True
                        mval = nval
                        mx = nx
                        my = ny

            local_max_val[cy, cx] = mval
            local_max_ind[cy, cx] = my * sx + mx

            if not changed:  # this pixel itself is the maximum in its neighborhood
                peak_found[cy, cx] = 1

    # find local peaks of all requested pixels
    cdef np.ndarray[np.int64_t, ndim=2, mode='c'] maxpath_np = np.zeros([1000, 2], dtype=np.int64)
    cdef long[:, ::1] maxpath = maxpath_np
    cdef long end_x, end_y, end_pos, end_ind, end_max_ind
    cdef long path_len = maxpath.shape[0]

    with nogil:
        for i in range(num_pixels):

            cx = px[i]

            if cx < 0 or cx >= sx:
                continue

            cy = py[i]

            if cy < 0 or cy >= sy:
                continue

            # initialize tracking trajectory
            end_pos = 0

            end_x = cx
            end_y = cy
            end_ind = cy * sx + cx
            end_max_ind = local_max_ind[end_y, end_x]

            maxpath[end_pos, 0] = end_x
            maxpath[end_pos, 1] = end_y

            while not peak_found[end_y, end_x]:

                # increment trajectory counter
                end_pos += 1

                # if overflow, increase size
                if end_pos >= path_len:
                    with gil:
                        maxpath_np.resize([path_len * 2, 2])
                    maxpath = maxpath_np
                    path_len = maxpath.shape[0]

                # add local max to trajectory
                end_ind = end_max_ind
                end_x = end_ind % sx
                end_y = end_ind / sx
                end_max_ind = local_max_ind[end_y, end_x]

                maxpath[end_pos, 0] = end_x
                maxpath[end_pos, 1] = end_y

            for i in range(end_pos+1):

                cx = maxpath[i, 0]
                cy = maxpath[i, 1]

                local_max_ind[cy, cx] = end_max_ind
                local_max_val[cy, cx] = local_max_val[end_y, end_x]
                peak_found[cy, cx] = 1

    return np.asarray(local_max_val), np.asarray(local_max_ind)



import numpy as np


def fit_poisson_mixture(im_input, mu=None, tol=0.1):
    """Generates a Poisson mixture model to fit pixel intensities for
    foreground/background masking.

    Takes as input an array or intensity image 'I' and optimizes a
    two-component poisson model describing foreground and background intensity
    models. This model can be used to describe the probability that a pixel
    comes from foreground versus background. The poisson distribution assumes
    discrete values and so is suitable for integral valued intensity images.
    Assumes that foreground intensities are lower (darker) than background.

    Parameters
    ----------
    im_input : array_like
        A hematoxylin intensity image obtained from ColorDeconvolution.
    mu : double
        Optional mean value of signal to optimize. Calculated from input if
        defined as 'None'. Default value = None.

    Returns
    -------
    thresh : double
        Optimal threshold for distinguishing foreground and background.
    im_fgnd : array_like
        An intensity image with values in the range [0, 1] representing
        foreground probabiities for each pixel.
    im_bgnd : array_like
        An intensity image with values in the range [0, 1] representing
        background probabiities for each pixel.

    References
    ----------
    .. [#] Y. Al-Kofahi et al "Improved Automatic Detection and Segmentation
       of Cell Nuclei in Histopathology Images" in IEEE Transactions on
       Biomedical Engineering,vol.57,no.4,pp.847-52, 2010.

    """
    import scipy as sp

    # check if intensity values in 'I' are integer type
    if not np.issubdtype(im_input.dtype, np.integer):
        msg = (
            'Inputs for Poisson mixture modeling should be integer'
            ' type'
        )
        raise TypeError(msg)

    # generate a small number for conditioning calculations
    Small = np.finfo(float).eps

    # generate histogram of inputs - assume range is 0, 255 (type uint8)
    H = np.histogram(np.ravel(im_input), bins=256, range=(0, 256))
    X = H[1]
    H = H[0].astype('float') / H[0].sum()

    if mu is None:
        mu = np.dot(X[0: -1], H)

    # calculate cumulative sum along histogram counts
    Cumulative = np.cumsum(H)
    CumProd = np.cumsum(np.multiply(X[0: -1], H))

    # determine cost at each possible threshold
    P0 = Cumulative
    P0[P0 <= 0] = Small
    P1 = 1 - Cumulative
    P1[P1 <= 0] = Small
    Mu0 = np.divide(CumProd, P0) + Small
    Mu1 = np.divide(CumProd[-1] - CumProd, P1) + Small
    Cost = mu - np.multiply(P0, np.log(P0) + np.multiply(Mu0, np.log(Mu0))) - \
        np.multiply(P1, np.log(P1) + np.multiply(Mu1, np.log(Mu1)))

    # identify minimum cost threshold
    thresh = X[np.argmin(Cost)]
    Mu0 = Mu0[np.argmin(Cost)]
    Mu1 = Mu1[np.argmin(Cost)]

    # build probability distribution of foreground intensity values
    Poisson = sp.stats.poisson(Mu0)
    Model = Poisson.pmf(np.arange(0, 256))
    im_fgnd = Model[im_input]
    Poisson = sp.stats.poisson(Mu1)
    Model = Poisson.pmf(np.arange(0, 256))
    im_bgnd = Model[im_input]

    # return threshold and foreground probabilities
    return thresh, im_fgnd, im_bgnd



import numpy as np


def merge_colinear(x, y):
    """Processes boundary coordinates in polyline with vertices X, Y to remove
    redundant colinear points. Polyline is not assumed to be open or closed.

    Parameters
    ----------
    x : array_like
        One dimensional array of horizontal boundary coordinates.
    y : array_like
        One dimensional array of vertical boundary coordinates.

    Returns
    -------
    xout : array_like
        X with colinear boundary points removed.
    yout : array_like
        Y with colinear boundary points removed.

    """
    # detect and delete points that are the same as the following point.
    Repeats = np.argwhere((np.diff(x) == 0) & (np.diff(y) == 0))
    xout = np.delete(x, Repeats)
    yout = np.delete(y, Repeats)

    # Calculating the slope for each transition could involve division by
    # zero, so we instead detect colinearity by noting that two non-zero
    # vectors are colinear if and only if the cross product of the vectors
    # is zero.  We convert to signed floats for two reasons: (1) in case
    # the inputs are unsigned values, (2) in case the cross product is
    # large.
    dX = np.diff(xout.astype(np.float64))
    dY = np.diff(yout.astype(np.float64))
    colinear = np.argwhere(dX[:-1] * dY[1:] == dX[1:] * dY[:-1]) + 1
    xout = np.delete(xout, colinear)
    yout = np.delete(yout, colinear)

    return xout, yout



import numpy as np


def graycomatrixext(im_input, im_roi_mask=None,
                    offsets=None, num_levels=None, gray_limits=None,
                    symmetric=False, normed=False, exclude_boundary=False):
    """Computes gray-level co-occurence matrix (GLCM) within a region of
    interest (ROI) of an image. GLCM is a 2D histogram/matrix containing the
    counts/probabilities of co-occuring intensity values at a given offset
    within an ROI of an image.

    Read the documentation to know the default values used for each of the
    optional parameter in different scenarios.

    Parameters
    ----------
    im_input : array_like
        Input single channel intensity image

    im_roi_mask : array_like, optional
        A binary mask specifying the region of interest within which
        to compute the GLCM. If not specified GLCM is computed for the
        the entire image.

        Default: None

    offsets : array_like, optional
        A (num_offsets, num_image_dims) array of offset vectors
        specifying the distance between the pixel-of-interest and
        its neighbor. Note that the first dimension corresponds to
        the rows.

        Because this offset is often expressed as an angle, the
        following table lists the offset values that specify common
        angles for a 2D image, given the pixel distance D.

        ===========  =============
        Angle (deg)  offset [y, x]
        ===========  =============
        0            [0 D]
        45           [-D D]
        90           [-D 0]
        135          [-D -D]
        ===========  =============

        Default
        - 1D: np.array([1])
        - 2D : numpy.array([ [1, 0], [0, 1], [1, 1], [1, -1] ])
        - 3D and higher: numpy.identity(num_image_dims)

    num_levels : unsigned int, optional
        An integer specifying the number of gray levels For example, if
        `NumLevels` is 8,  the intensity values of the input image are
        scaled so they are integers between 1 and 8.  The number of gray
        levels determines the size of the gray-level co-occurrence matrix.

        Default: 2 for binary/logical image, 32 for numeric image

    gray_limits : array_like, optional
        A two-element array specifying the desired input intensity range.
        Intensity values in the input image will be clipped into this range.

        Default: [0, 1] for boolean-valued image, [0, 255] for integer-valued
        image, and [0.0, 1.0] for-real valued image

    symmetric : bool, optional
        A boolean value that specifies whether or not the ordering of values
        in pixel pairs is considered while creating the GLCM matrix.

        For example, if `Symmetric` is True, then while calculating the
        number of times the value 1 is adjacent to the value 2, both
        1,2 and 2,1 pairings are counted. GLCM created in this way is
        symmetric across its diagonal.

        Default: False

    normed : bool, optional
        A boolean value specifying whether or not to normalize glcm.

        Default: False

    exclude_boundary : bool, optional
        Specifies whether or not to exclude a pixel-pair if the
        neighboring pixel in the pair is outside `im_roi_mask`.
        Has an effect only when `im_roi_mask` is specified.

        Default: False

    Returns
    -------
    glcm : array_like
        num_levels x num_levels x num_offsets array containing the GLCM
        for each offset.

    References
    ----------
    .. [#] Haralick, R.M., K. Shanmugan, and I. Dinstein, "Textural Features
       for Image Classification", IEEE Transactions on Systems, Man, and
       Cybernetics, Vol. SMC-3, 1973, pp. 610-621.

    .. [#] Haralick, R.M., and L.G. Shapiro. Computer and Robot Vision:
       Vol. 1, Addison-Wesley, 1992, p. 459.

    """
    num_dims = len(im_input.shape)

    # roi mask
    if im_roi_mask is None:

        # compute glcm for whole input image
        im_roi_mask = np.ones_like(im_input, dtype='bool')

    if im_input.shape != im_roi_mask.shape:
        msg = 'size mismatch between input image and roi mask'
        raise ValueError(msg)

    # gray_limits
    if gray_limits is None:

        gray_limits = _default_gray_limits(im_input)

    assert len(gray_limits) == 2
    assert gray_limits[0] < gray_limits[1]

    # num_levels
    if num_levels is None:

        num_levels = _default_num_levels(im_input)

    # offsets
    if offsets is None:

        # set default offset value
        offsets = _default_offsets(im_input)

    else:

        # check sanity
        if offsets.shape[1] != num_dims:
            msg = 'Dimension mismatch between input image and offsets'
            raise ValueError(
                msg,
            )

    num_offsets = offsets.shape[0]

    # scale input intensity image
    im_input = im_input.astype('float')
    im_input -= gray_limits[0]
    im_input /= float(gray_limits[1] - gray_limits[0])
    im_input *= (num_levels - 1)
    im_input = np.round(im_input).astype('int')

    # compute glcm for each offset
    glcm = np.zeros((num_levels, num_levels, num_offsets))

    im_input_flat = np.ravel(im_input)

    im_roi_mask_flat = np.ravel(im_roi_mask)

    roi_coord_ind = np.nonzero(im_roi_mask)

    roi_lin_ind = np.ravel_multi_index(roi_coord_ind, im_roi_mask.shape)

    for i in range(num_offsets):

        # compute indices of neighboring pixels by applying the offset
        neigh_coord_ind = [None] * len(roi_coord_ind)

        for j in range(num_dims):
            neigh_coord_ind[j] = roi_coord_ind[j] + offsets[i, j]

        # throw out pixels with invalid neighbors
        neigh_valid = np.ones_like(neigh_coord_ind[0], dtype='bool')

        for j in range(num_dims):

            neigh_valid[neigh_coord_ind[j] < 0] = False
            neigh_valid[neigh_coord_ind[j] >= im_roi_mask.shape[j]] = False

        for j in range(num_dims):
            neigh_coord_ind[j] = np.compress(neigh_valid, neigh_coord_ind[j],
                                             axis=0).astype(np.int64)

        neigh_lin_ind = np.ravel_multi_index(neigh_coord_ind,
                                             im_roi_mask.shape)

        if exclude_boundary:
            neigh_valid[im_roi_mask_flat[neigh_lin_ind] == 0] = False
            neigh_lin_ind = np.compress(neigh_valid, neigh_lin_ind, axis=0)

        valid_roi_lin_ind = np.compress(neigh_valid, roi_lin_ind, axis=0)

        # get intensities of pixel pairs which become coord indices in glcm
        p1 = np.take(im_input_flat, valid_roi_lin_ind, axis=0)
        p2 = np.take(im_input_flat, neigh_lin_ind, axis=0)

        # convert pixel-pair values to linear indices in glcm
        pind = np.ravel_multi_index((p1, p2), glcm.shape[:2])

        # find unique linear indices and their counts
        pind, pcount = np.unique(pind, return_counts=True)

        # put count of each linear index in glcm
        cur_glcm = np.zeros((num_levels, num_levels))
        cur_glcm_flat = np.ravel(cur_glcm)
        cur_glcm_flat[pind] = pcount

        # symmetricize if asked for
        if symmetric:
            cur_glcm += cur_glcm.T

        # normalize if asked
        if normed and cur_glcm.sum():
            cur_glcm /= cur_glcm.sum()

        glcm[:, :, i] = cur_glcm

    return glcm


def _default_gray_limits(im_input):

    assert isinstance(im_input, np.ndarray)

    if np.issubdtype(im_input.dtype, np.bool_):

        gray_limits = [0, 1]

    elif np.issubdtype(im_input.dtype, np.integer):

        gray_limits = [0, 255]

    elif np.issubdtype(im_input.dtype, np.floating):

        gray_limits = [0.0, 1.0]

    else:

        msg = 'The type of the argument im_input is invalid'
        raise ValueError(msg)

    return gray_limits


def _default_num_levels(im_input):

    assert isinstance(im_input, np.ndarray)

    if np.issubdtype(im_input.dtype, np.bool_):

        num_levels = 2

    elif np.issubdtype(im_input.dtype, np.number):

        num_levels = 32

    else:

        msg = 'The type of the argument im_input is invalid'
        raise ValueError(msg)

    return num_levels


def _default_offsets(im_input):

    num_dims = len(im_input.shape)

    if num_dims == 2:

        offsets = np.array([
            [0, 1], [1, 0], [1, 1], [1, -1],
        ])

    else:

        # TODO: need to come up with a better strategy for 3D and higher
        offsets = np.identity(num_dims)

    return offsets



"""
This package contains functions that implement commonly used level-set based
methods for segmenting objects/regions in images.
"""

# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .chan_vese import chan_vese
from .reg_edge import reg_edge

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'chan_vese',
    'reg_edge',
)



import numpy as np

from .rgb_to_lab import _lms2lab, _rgb2lms

# Define conversion matrices
_lms2rgb = np.linalg.inv(_rgb2lms)
_lab2lms = np.linalg.inv(_lms2lab)


def lab_to_rgb(im_lab):
    """Transforms an image from LAB to RGB color space

    Parameters
    ----------
    im_lab : array_like
        An image in LAB color space

    Returns
    -------
    im_rgb : array_like
        The RGB representation of the input image 'im_lab'.

    See Also
    --------
    histomicstk.preprocessing.color_conversion.rgb_to_lab,
    histomicstk.preprocessing.color_normalization.reinhard

    References
    ----------
    .. [#] D. Ruderman, T. Cronin, and C. Chiao, "Statistics of cone
       responses to natural images: implications for visual coding,"
       J. Opt. Soc. Am. A 15, 2036-2045 (1998).

    """
    # get input image dimensions
    m = im_lab.shape[0]
    n = im_lab.shape[1]

    # calculate im_lms values from LAB
    im_lab = np.reshape(im_lab, (m * n, 3))
    im_lms = np.dot(_lab2lms, np.transpose(im_lab))

    # calculate RGB values from im_lms
    im_lms = np.exp(im_lms)
    im_lms[im_lms == np.spacing(1)] = 0

    im_rgb = np.dot(_lms2rgb, im_lms)

    # reshape to 3-channel image
    im_rgb = np.reshape(im_rgb.transpose(), (m, n, 3))

    return im_rgb



"""
This package contains implementation of methods to deconvolve or separate.

... the stains of histopathology images.
"""
# make functions available at the package level using shadow imports
# since we mostly have one function per file
from . import stain_color_map as _stain_color_map
from .color_convolution import color_convolution
# must be imported after ComplementStainMatrix
from .color_deconvolution import (_reorder_stains, color_deconvolution,
                                  color_deconvolution_routine,
                                  stain_unmixing_routine)
from .complement_stain_matrix import complement_stain_matrix
from .find_stain_index import find_stain_index
from .rgb_separate_stains_macenko_pca import rgb_separate_stains_macenko_pca
from .rgb_separate_stains_xu_snmf import rgb_separate_stains_xu_snmf
from .separate_stains_macenko_pca import separate_stains_macenko_pca
from .separate_stains_xu_snmf import separate_stains_xu_snmf

#: A dictionary of names for reference stain vectors
stain_color_map = _stain_color_map.stain_color_map

# list out things that are available for public use
__all__ = (

    # functions, classes, and constants of this package
    'color_convolution',
    'color_deconvolution',
    'stain_unmixing_routine',
    'color_deconvolution_routine',
    'complement_stain_matrix',
    'find_stain_index',
    'separate_stains_macenko_pca',
    'separate_stains_xu_snmf',
    'rgb_separate_stains_macenko_pca',
    'rgb_separate_stains_xu_snmf',
    'stain_color_map',
    '_reorder_stains',
)



import numpy as np


def rag_color(adj_mat):
    """Generates a coloring of an adjacency graph using the sequential coloring
    algorithm. Used to bin regions from a label image into a small number of
    independent groups that can be processed separately with algorithms like
    multi-label graph cuts or individual active contours. The rationale is to
    color adjacent objects with distinct colors so that their contours can be
    co-evolved.

    Parameters
    ----------
    adj_mat : array_like
        A binary matrix of size N x N, where N is the number of objects in
        Label. A value of 'True' at adj_mat(i,j) indicates that objects 'i'
        and 'j' are neighbors. Does not contain entries for background objects.

    Returns
    -------
    Colors : array_like
        A list of colors for the objects encoded in 'adj_mat'. No two objects
        that are connected in 'adj_mat' will share the same color.

    See Also
    --------
    histomicstk.segmentation.rag,
    histomicstk.segmentation.rag_add_layer

    """
    # initialize colors and color count
    Colors = np.zeros((adj_mat.shape[0], 1), dtype=int)
    Colors[0] = 1
    ColorCount = 1

    # iterate over remaining nodes in order, finding legal coloring
    for i in range(1, adj_mat.shape[0]):

        # get indices neighbors of node 'i'
        Neighbors = np.nonzero(adj_mat[i, ])[0].flatten()

        if Neighbors.size > 0:

            # get colors of neighbors
            NeighborColors = Colors[Neighbors]
            NeighborColors = NeighborColors[np.nonzero(NeighborColors)]

            # check if neighbors have been labeled
            if NeighborColors.size > 0:

                # find lowest legal color of node 'i'
                Reference = set(range(1, ColorCount + 1))
                Diff = Reference.difference(set(NeighborColors))
                if len(Diff) == 0:
                    ColorCount += 1
                    Colors[i] = ColorCount
                else:
                    Colors[i] = min(Diff)
            else:

                # no other neighbors have been labeled yet - set value = 1
                Colors[i] = 1

        else:  # object is an island - no neighbors

            # set to base color
            Colors[i] = 1

    return Colors.flatten()



import numpy as np

cimport cython
cimport numpy as np
from libc.math cimport M_PI, cos, round, sin
from libcpp.vector cimport vector


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _trace_object_boundaries_cython(long[:, :] im_label not None, long connectivity, long x_start, long y_start, double max_length):

    cdef long nrows = im_label.shape[0]
    cdef long ncols = im_label.shape[1]

    cdef int flag = 0

    cdef long[:, :] im_label_270 = _rot90(im_label)
    cdef long[:, :] im_label_180 = _rot90(im_label_270)
    cdef long[:, :] im_label_90 = _rot90(im_label_180)

    cdef long i, j, t
    with nogil:

        # find starting x and y points if not defined
        if (x_start == -1) & (y_start == -1):

          for i in range(nrows):
            for j in range(ncols):

              if (im_label[i, j] > 0) & (flag == 0):
                # check if the number of points is one
                if ((im_label[i, j+1] == 0) & (im_label[i+1, j] == 0) & (im_label[i+1, j+1] == 0) & (im_label[i-1, j+1] == 0)) == False:
                  x_start = j
                  y_start = i
                  flag = 1
                  break
            if flag == 1:
                break

    bx = []
    by = []

    if connectivity == 4:
        bx, by = _isbf(
            im_label, im_label_90, im_label_180, im_label_270,
            x_start, y_start, max_length);

    else:
        bx, by = _moore(
            im_label, im_label_90, im_label_180, im_label_270,
            x_start, y_start, max_length);

    return np.asarray(bx), np.asarray(by)


cdef long[:, :] _rot90(long[:, :] input):

    cdef long nrows = input.shape[0]
    cdef long ncols = input.shape[1]

    cdef long[:, :] output = np.zeros([ncols, nrows], dtype=int)

    cdef long i, j

    for i in range(nrows):
        for j in range(ncols):
          output[j, nrows-1-i] = input[i, j]

    return output


def _moore(long[:, :] mask, long[:, :] mask_90, long[:, :] mask_180, long[:, :] mask_270, int x_start, int y_start, float max_length):

    cdef long nrows = mask.shape[0]
    cdef long ncols = mask.shape[1]

    # initialize boundary vector
    cdef vector[long] list_bx
    cdef vector[long] list_by

    # push the first x and y points
    list_bx.push_back(x_start);
    list_by.push_back(y_start);

    # check degenerate case where mask contains 1 pixel
    cdef long sum = np.sum(mask)

    # set size of boundary points: the size of X is equal to the size of Y
    cdef long size_boundary

    cdef long[:, :] h
    cdef long a, b, x, y

    # initialize cX and cY which indicate directions for each Moore
    cdef long cX
    cdef long cY

    # set sin and cos
    cdef double s, c, p, q, ci, cj

    # initialize default direction
    cdef long DX = 1
    cdef long DY = 0

    # set the number of rows and cols for ISBF
    cdef long row_isbf = 3
    cdef long col_isbf = 3

    cdef double angle

    # define clockwise ordered indices
    cdef vector[long] row = [2, 1, 0, 0, 0, 1, 2, 2]
    cdef vector[long] col = [0, 0, 0, 1, 2, 2, 2, 1]
    cdef vector[long] dX = [-1, 0, 0, 1, 1, 0, 0, -1]
    cdef vector[long] dY = [0, -1, -1, 0, 0, 1, 1, 0]
    cdef vector[long] oX = [-1, -1, -1, 0, 1, 1, 1, 0]
    cdef vector[long] oY = [1, 0, -1, -1, -1, 0, 1, 1]

    cdef long move = 0
    cdef int is_moore = 0

    cdef long fx1, fx2, fy1, fy2
    cdef long lx1, lx2, ly1, ly2

    cdef long i, j

    if sum > 1:

      # loop until true
      while(True):

        h = np.zeros((row_isbf, col_isbf), dtype=int)

        with nogil:
            with cython.boundscheck(False):
                # initialize a and b which are indices of moore
                a = 0
                b = 0

                # get length of the current linked list
                size_boundary = list_bx.size()

                x = list_bx[size_boundary-1]
                y = list_by[size_boundary-1]

                if (DX == 1) & (DY == 0):
                  for i in range(ncols-x-2, ncols-x+1):
                    for j in range(y-1, y+2):
                        h[a, b] = mask_90[i, j]
                        b = b + 1
                    b = 0
                    a = a + 1
                  angle = M_PI/2

                elif (DX == 0) & (DY == -1):
                  for i in range(y-1, y+2):
                    for j in range(x-1, x+2):
                        h[a, b] = mask[i, j]
                        b = b + 1
                    b = 0
                    a = a + 1
                  angle = 0

                elif (DX == -1) & (DY == 0):
                  for i in range(x-1, x+2):
                    for j in range(nrows-y-2, nrows-y+1):
                        h[a, b] = mask_270[i, j]
                        b = b + 1
                    b = 0
                    a = a + 1
                  angle = 3*M_PI/2

                else:
                  for i in range(nrows-y-2, nrows-y+1):
                    for j in range(ncols-x-2, ncols-x+1):
                        h[a, b] = mask_180[i, j]
                        b = b + 1
                    b = 0
                    a = a + 1
                  angle = M_PI

                move = 0
                is_moore = 0

                for i in range(8):
                  if is_moore == 0:
                    if h[row[i], col[i]] == 1:
                        move = i
                        is_moore = 1

                cX = oX[move]
                cY = oY[move]
                DX = dX[move]
                DY = dY[move]

                # transform points by incoming directions and add to contours
                s = sin(angle)
                c = cos(angle)

                p = c*cX - s*cY
                q = s*cX + c*cY

        list_bx.push_back(list_bx.back()+int(round(p)))
        list_by.push_back(list_by.back()+int(round(q)))

        with nogil:
            ci = c*DX-s*DY
            cj = s*DX+c*DY

            DX = int(round(ci))
            DY = int(round(cj))

            # get length of the current linked list
            size_boundary = list_bx.size()

            if size_boundary > 3:

              fx1 = list_bx[0]
              fx2 = list_bx[1]
              fy1 = list_by[0]
              fy2 = list_by[1]
              lx1 = list_bx[size_boundary-1]
              ly1 = list_by[size_boundary-1]
              lx2 = list_bx[size_boundary-2]
              ly2 = list_by[size_boundary-2]

              # check if the first and the last x and y are equal
              if (size_boundary > max_length) | \
              ((lx1 == fx2)&(lx2 == fx1)&(ly1 == fy2)&(ly2 == fy1)):
                  # remove the last element
                  list_bx.pop_back()
                  list_by.pop_back()
                  break

    return list_bx, list_by


def _isbf(long[:, :] mask, long[:, :] mask_90, long[:, :] mask_180, long[:, :] mask_270, int x_start, int y_start, float max_length):

    cdef long nrows = mask.shape[0]
    cdef long ncols = mask.shape[1]

    # initialize boundary vector
    cdef vector[long] list_bx
    cdef vector[long] list_by

    # set default direction
    cdef long DX = 1
    cdef long DY = 0

    # set the number of rows and cols for ISBF
    cdef long row_isbf = 3
    cdef long col_isbf = 2

    cdef double angle

    # set size of boundary points: the size of X is equal to the size of Y
    cdef long size_boundary

    cdef long[:, :] h
    cdef long a, b, x, y

    # initialize cX and cY which indicate directions for each ISBF
    cdef vector[int] cX
    cdef vector[int] cY

    # set sin and cos
    cdef double s, c, cx, cy, ci, cj

    cdef long fx1, fx2, fy1, fy2
    cdef long lx1, lx2, ly1, ly2
    cdef long lx3, lx4, ly3, ly4

    cdef long i, j, t

    # check degenerate case where mask contains 1 pixel
    cdef long sum = np.sum(mask)

    if sum > 1:
        with nogil:
            # push the first x and y points
            list_bx.push_back(x_start);
            list_by.push_back(y_start);

        while(True):

            h = np.zeros((row_isbf, col_isbf), dtype=int)

            with nogil:
                with cython.boundscheck(False):
                    # initialize a and b which are indices of ISBF
                    a = 0
                    b = 0

                    # get length of the current linked list
                    size_boundary = list_bx.size()

                    x = list_bx[size_boundary-1]
                    y = list_by[size_boundary-1]

                    if (DX == 1) & (DY == 0):
                        for i in range(ncols-x-2, ncols-x+1):
                            for j in range(y-1, y+1):
                                h[a, b] = mask_90[i, j]
                                b = b + 1
                            b = 0
                            a = a + 1
                        angle = M_PI/2

                    elif (DX == 0) & (DY == -1):
                        for i in range(y-1, y+2):
                            for j in range(x-1, x+1):
                                h[a, b] = mask[i, j]
                                b = b + 1
                            b = 0
                            a = a + 1
                        angle = 0

                    elif (DX == -1) & (DY == 0):
                        for i in range(x-1, x+2):
                            for j in range(nrows-y-2, nrows-y):
                                h[a, b] = mask_270[i, j]
                                b = b + 1
                            b = 0
                            a = a + 1
                        angle = 3*M_PI/2

                    else:
                        for i in range(nrows-y-2, nrows-y+1):
                            for j in range(ncols-x-2, ncols-x):
                                h[a, b] = mask_180[i, j]
                                b = b + 1
                            b = 0
                            a = a + 1
                        angle = M_PI

                    cX = vector[int](1)
                    cY = vector[int](1)

                    if h[1, 0] == 1:
                        # 'left' neighbor
                        cX[0] = -1
                        cY[0] = 0
                        DX = -1
                        DY = 0
                    else:
                        if (h[2][0] == 1) & (h[2][1] != 1):
                            # inner-outer corner at left-rear
                            cX[0] = -1
                            cY[0] = 1
                            DX = 0
                            DY = 1
                        else:
                            if h[0, 0] == 1:
                                if h[0, 1] == 1:
                                    # inner corner at front
                                    cX[0] = 0
                                    cY[0] = -1
                                    cX.push_back(-1)
                                    cY.push_back(0)
                                    DX = 0
                                    DY = -1
                                else:
                                    # inner-outer corner at front-left
                                    cX[0] = -1
                                    cY[0] = -1
                                    DX = 0
                                    DY = -1
                            elif h[0, 1] == 1:
                                # front neighbor
                                cX[0] = 0
                                cY[0] = -1
                                DX = 1
                                DY = 0
                            else:
                                # outer corner
                                DX = 0
                                DY = 1

                    # transform points by incoming directions and add to contours
                    s = sin(angle)
                    c = cos(angle)

                    if (cX[0]!=0) | (cY[0]!=0):

                        for t in range(cX.size()):

                            cx = c*cX[t] - s*cY[t]
                            cy = s*cX[t] + c*cY[t]

                            with gil:
                                list_bx.push_back(list_bx.back()+int(round(cx)))
                                list_by.push_back(list_by.back()+int(round(cy)))

                    ci = c*DX - s*DY
                    cj = s*DX + c*DY

                    DX = int(round(ci))
                    DY = int(round(cj))

                    # get length of the current linked list
                    size_boundary = list_bx.size()

                    if size_boundary > 3:

                        fx1 = list_bx[0]
                        fx2 = list_bx[1]
                        fy1 = list_by[0]
                        fy2 = list_by[1]
                        lx1 = list_bx[size_boundary-1]
                        ly1 = list_by[size_boundary-1]
                        lx2 = list_bx[size_boundary-2]
                        ly2 = list_by[size_boundary-2]
                        lx3 = list_bx[size_boundary-3]
                        ly3 = list_by[size_boundary-3]
                        lx4 = list_bx[size_boundary-4]
                        ly4 = list_by[size_boundary-4]

                        # check if the first and the last x and y are equal
                        if (size_boundary > max_length) | \
                                ((lx1 == fx2)&(lx2 == fx1)&(ly1 == fy2)&(ly2 == fy1)):
                            # remove the last element
                            list_bx.pop_back()
                            list_by.pop_back()
                            break
                        if cX.size() == 2:
                            if (lx2 == fx2)&(lx3 == fx1)&(ly2 == fy2)&(ly3 == fy1):
                                list_bx.pop_back()
                                list_by.pop_back()
                                list_bx.pop_back()
                                list_by.pop_back()
                                break
                        # detect cycle
                        if (lx1 == lx3)&(ly1 == ly3)&(lx2 == lx4)&(ly2 == ly4):
                            list_bx.pop_back()
                            list_by.pop_back()
                            list_bx.pop_back()
                            list_by.pop_back()
                            # change direction from M_PI to 3*M_PI/2
                            if (DX == 0) & (DY == 1):
                                DX = -1
                                DY = 0
                            # from M_PI/2 to M_PI
                            elif (DX == 1) & (DY == 0):
                                DX = 0
                                DY = 1
                            # from 0 to M_PI/2
                            elif (DX == 0) & (DY == -1):
                                DX = 1
                                DY = 0
                            else:
                                DX = 0
                                DY = -1

    return list_bx, list_by



"""Functions to correct non-uniform staining issues."""

# make functions available at the package level using shadow imports
# since we mostly have one function per file
from .background_intensity import background_intensity
from .deconvolution_based_normalization import \
    deconvolution_based_normalization
from .reinhard import reinhard
from .reinhard_stats import reinhard_stats, reinhard_stats_rgb

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'background_intensity',
    'reinhard',
    'reinhard_stats',
    'reinhard_stats_rgb',
    'deconvolution_based_normalization',
)



import numpy as np


def compute_gradient_features(im_label, im_intensity,
                              num_hist_bins=10, rprops=None):
    """Calculates gradient features from an intensity image.

    Parameters
    ----------
    im_label : array_like
        A labeled mask image wherein intensity of a pixel is the ID of the
        object it belongs to. Non-zero values are considered to be foreground
        objects.
    im_intensity : array_like
        Intensity image
    num_hist_bins: int, optional
        Number of bins used to computed the gradient histogram of an object.
        Histogram is used to energy and entropy features. Default is 10.
    rprops : output of skimage.measure.regionprops, optional
        rprops = skimage.measure.regionprops( im_label ). If rprops is not
        passed then it will be computed inside which will increase the
        computation time.

    Returns
    -------
    fdata: pandas.DataFrame
        A pandas dataframe containing the gradient features listed below for
        each object/label.

    Notes
    -----
    List of gradient features computed by this function:

    Gradient.Mag.Mean : float
        Mean of gradient data.

    Gradient.Mag.Std : float
        Standard deviation of gradient data.

    Gradient.Mag.Skewness : float
        Skewness of gradient data. Value is 0 when all values are equal.

    Gradient.Mag.Kurtosis : float
        Kurtosis of gradient data. Value is -3 when all values are equal.

    Gradient.Mag.HistEnergy : float
        Energy of the gradient magnitude histogram of object pixels

    Gradient.Mag.HistEnergy : float
        Entropy of the gradient magnitude histogram of object pixels.

    Gradient.Canny.Sum : float
        Sum of canny filtered gradient data.

    Gradient.Canny.Mean : float
        Mean of canny filtered gradient data.

    References
    ----------
    .. [#] Daniel Zwillinger and Stephen Kokoska. "CRC standard probability
       and statistics tables and formulae," Crc Press, 1999.

    """
    import pandas as pd
    import scipy.stats
    from skimage.feature import canny
    from skimage.measure import regionprops

    # List of feature names
    feature_list = [
        'Gradient.Mag.Mean',
        'Gradient.Mag.Std',
        'Gradient.Mag.Skewness',
        'Gradient.Mag.Kurtosis',
        'Gradient.Mag.HistEntropy',
        'Gradient.Mag.HistEnergy',
        'Gradient.Canny.Sum',
        'Gradient.Canny.Mean',
    ]

    # compute object properties if not provided
    if rprops is None:
        rprops = regionprops(im_label)

    # create pandas data frame containing the features for each object
    numFeatures = len(feature_list)
    numLabels = len(rprops)
    fdata = pd.DataFrame(np.zeros((numLabels, numFeatures)),
                         columns=feature_list)

    Gx, Gy = np.gradient(im_intensity)
    diffG = np.sqrt(Gx**2 + Gy**2)
    cannyG = canny(im_intensity)

    for i in range(numLabels):
        if rprops[i] is None:
            continue

        # get gradients of object pixels
        pixelGradients = np.sort(
            diffG[rprops[i].coords[:, 0], rprops[i].coords[:, 1]],
        )

        # compute mean
        fdata.at[i, 'Gradient.Mag.Mean'] = np.mean(pixelGradients)

        # compute standard deviation
        fdata.at[i, 'Gradient.Mag.Std'] = np.std(pixelGradients)

        # compute skewness
        fdata.at[i, 'Gradient.Mag.Skewness'] = scipy.stats.skew(pixelGradients)

        # compute kurtosis
        fdata.at[i, 'Gradient.Mag.Kurtosis'] = \
            scipy.stats.kurtosis(pixelGradients)

        # compute intensity histogram
        hist, bins = np.histogram(pixelGradients, bins=num_hist_bins)
        prob = hist / np.sum(hist, dtype=np.float32)

        # compute entropy
        fdata.at[i, 'Gradient.Mag.HistEntropy'] = scipy.stats.entropy(prob)

        # compute energy
        fdata.at[i, 'Gradient.Mag.HistEnergy'] = np.sum(prob**2)

        bw_canny = cannyG[rprops[i].coords[:, 0], rprops[i].coords[:, 1]]
        canny_sum = np.sum(bw_canny).astype('float')

        fdata.at[i, 'Gradient.Canny.Sum'] = canny_sum

        fdata.at[i, 'Gradient.Canny.Mean'] = canny_sum / len(pixelGradients)

    return fdata



import numpy as np

import histomicstk.features as htk_features
import histomicstk.preprocessing.color_deconvolution as htk_cdeconv
import histomicstk.preprocessing.color_normalization as htk_cnorm
import histomicstk.segmentation.label as htk_seg_label
import histomicstk.segmentation.nuclear as htk_nuclear
from histomicstk.cli import utils as cli_utils


def detect_tile_nuclei(tile_info, args, src_mu_lab=None,
                       src_sigma_lab=None, invert_image=False,
                       default_img_inversion=False, return_fdata=False):
    """
    Detect nuclei within a tile image and generate annotations.

    Args:
    ----
        tile_info (dict): Information about the tile image.
        args: Arguments from the cli.
        src_mu_lab: Source mean in LAB color space.
        src_sigma_lab: Source standard deviation in LAB color space.
        invert_image (bool): Invert the image.
        default_img_inversion (bool): Apply default image inversion.
        return_fdata (bool): Return computed features.

    Returns:
    -------
        nuclei_annot_list (list): List of nuclei annotations.
        fdata: Computed nuclei features.

    """
    # Flags
    single_channel = False

    # get tile image & check number of channels
    single_channel = len(tile_info['tile'].shape) <= 2 or tile_info['tile'].shape[2] == 1
    if single_channel:
        im_tile = np.dstack((tile_info['tile'], tile_info['tile'], tile_info['tile']))
        if default_img_inversion:
            invert_image = True
    else:
        im_tile = tile_info['tile'][:, :, :3]

    # perform image inversion
    if invert_image:
        im_tile = (np.iinfo(im_tile.dtype).max
                   if im_tile.dtype.kind == 'u'
                   else np.max(im_tile)) - im_tile

    # perform color normalization
    im_nmzd = htk_cnorm.reinhard(im_tile,
                                 args.reference_mu_lab,
                                 args.reference_std_lab,
                                 src_mu=src_mu_lab,
                                 src_sigma=src_sigma_lab)

    # perform color decovolution
    w = cli_utils.get_stain_matrix(args)

    # perform deconvolution
    im_stains = htk_cdeconv.color_deconvolution(im_nmzd, w).Stains
    im_nuclei_stain = im_stains[:, :, 0].astype(float)

    # segment nuclear foreground
    im_nuclei_fgnd_mask = im_nuclei_stain < args.foreground_threshold

    # segment nuclei
    im_nuclei_seg_mask = htk_nuclear.detect_nuclei_kofahi(
        im_nuclei_stain,
        im_nuclei_fgnd_mask,
        args.min_radius,
        args.max_radius,
        args.min_nucleus_area,
        args.local_max_search_radius,
    )

    # Delete border nuclei
    if args.ignore_border_nuclei is True:

        im_nuclei_seg_mask = htk_seg_label.delete_border(im_nuclei_seg_mask)

    # Delete overlapping border nuclei
    if any(tile_info['tile_overlap'].values()) > 0:

        im_nuclei_seg_mask = htk_seg_label.delete_overlap(
            im_nuclei_seg_mask, overlap_info=tile_info['tile_overlap'])

    # generate nuclei annotations
    nuclei_annot_list = []

    flag_nuclei_found = np.any(im_nuclei_seg_mask)

    # compute nuclei features and nuclei_annot_list
    fdata = None

    if flag_nuclei_found:
        if args.nuclei_annotation_format:
            format = args.nuclei_annotation_format
            if args.nuclei_annotation_format == 'bbox' and - \
                    args.remove_overlapping_nuclei_segmentation:
                format = 'boundary'

        if return_fdata:
            if args.cytoplasm_features:
                im_cytoplasm_stain = im_stains[:, :, 1].astype(float)
            else:
                im_cytoplasm_stain = None
            # Generate features and nuclei annotation simultaneously
            fdata, nuclei_annot_list = htk_features.compute_nuclei_features(
                im_nuclei_seg_mask,
                im_nuclei_stain.astype(np.uint8), im_cytoplasm_stain.astype(np.uint8),
                fsd_bnd_pts=args.fsd_bnd_pts,
                fsd_freq_bins=args.fsd_freq_bins,
                cyto_width=args.cyto_width,
                num_glcm_levels=args.num_glcm_levels,
                morphometry_features_flag=args.morphometry_features,
                fsd_features_flag=args.fsd_features,
                intensity_features_flag=args.intensity_features,
                gradient_features_flag=args.gradient_features,
                tile_info=tile_info,
                im_nuclei_seg_mask=im_nuclei_seg_mask,
                format=format,
                return_nuclei_annotation=True,
            )
            fdata.columns = ['Feature.' + col for col in fdata.columns]
        else:
            nuclei_annot_list = cli_utils.create_tile_nuclei_annotations(
                im_nuclei_seg_mask, tile_info, format)

    if return_fdata:
        return nuclei_annot_list, fdata

    return nuclei_annot_list



"""Placeholder."""
import numpy as np

from histomicstk.preprocessing import color_conversion


def reinhard(
        im_src, target_mu, target_sigma, src_mu=None, src_sigma=None,
        mask_out=None):
    """Perform Reinhard color normalization.

    Transform the color characteristics of an image to a desired standard.
    The standard is defined by the mean and standard deviations of the target
    image in LAB color space defined by Ruderman. The input image is converted
    to Ruderman's LAB space, the LAB channels are each centered and scaled to
    zero-mean unit variance, and then rescaled and shifted to match the target
    image statistics. If the LAB statistics for the input image are provided
    (`src_mu` and `src_sigma`) then these will be used for normalization,
    otherwise they will be derived from the input image `im_src`.

    Parameters
    ----------
    im_src : array_like
        An RGB image

    target_mu : array_like
        A 3-element array containing the means of the target image channels
        in LAB color space.

    target_sigma : array_like
        A 3-element array containing the standard deviations of the target
        image channels in LAB color space.

    src_mu : array_like, optional
        A 3-element array containing the means of the source image channels in
        LAB color space. Used with reinhard_stats for uniform normalization of
        tiles from a slide.

    src_sigma : array, optional
        A 3-element array containing the standard deviations of the source
        image channels in LAB color space. Used with reinhard_stats for
        uniform normalization of tiles tiles from a slide.

    mask_out : array_like, default is None
        if not None, should be (m, n) boolean numpy array.
        This method uses numpy masked array functionality to only use
        non-masked areas in calculations. This is relevant because elements
        like blood, sharpie marker, white space, etc would throw off the
        reinhard normalization by affecting the mean and stdev. Ideally, you
        want to exclude these elements from both the target image (from which
        you calculate target_mu and target_sigma) and from the source image
        to be normalized.

    Returns
    -------
    im_normalized : array_like
        Color Normalized RGB image

    See Also
    --------
    histomicstk.preprocessing.color_conversion.rgb_to_lab,
    histomicstk.preprocessing.color_conversion.lab_to_rgb

    References
    ----------
    .. [#] E. Reinhard, M. Adhikhmin, B. Gooch, P. Shirley, "Color transfer
       between images," in IEEE Computer Graphics and Applications, vol.21,
       no.5,pp.34-41, 2001.
    .. [#] D. Ruderman, T. Cronin, and C. Chiao, "Statistics of cone responses
       to natural images: implications for visual coding," J. Opt. Soc. Am. A
       vol.15, pp.2036-2045, 1998.

    """
    # convert input image to LAB color space
    im_lab = color_conversion.rgb_to_lab(im_src)

    # mask out irrelevant tissue / whitespace / etc
    if mask_out is not None:
        mask_out = mask_out[..., None]
        im_lab = np.ma.masked_array(
            im_lab, mask=np.tile(mask_out, (1, 1, 3)))

    # calculate src_mu and src_sigma if either is not provided
    if (src_mu is None) or (src_sigma is None):
        src_mu = [im_lab[..., i].mean() for i in range(3)]
        src_sigma = [im_lab[..., i].std() for i in range(3)]

    # scale to unit variance
    for i in range(3):
        im_lab[:, :, i] = (im_lab[:, :, i] - src_mu[i]) / src_sigma[i]

    # rescale and recenter to match target statistics
    for i in range(3):
        im_lab[:, :, i] = im_lab[:, :, i] * target_sigma[i] + target_mu[i]

    # convert back to RGB colorspace
    im_normalized = color_conversion.lab_to_rgb(im_lab)
    im_normalized[im_normalized > 255] = 255
    im_normalized[im_normalized < 0] = 0

    # return masked values and reconstruct unmasked LAB image
    if mask_out is not None:
        im_normalized = im_normalized.data
        for i in range(3):
            original = im_src[:, :, i].copy()
            new = im_normalized[:, :, i].copy()
            original[np.not_equal(mask_out[:, :, 0], True)] = 0
            new[mask_out[:, :, 0]] = 0
            im_normalized[:, :, i] = new + original
    im_normalized = im_normalized.astype(np.uint8)

    return im_normalized



"""
This package contains utility functions to convert images between different
color spaces.
"""
from .lab_mean_std import lab_mean_std  # after rgb_to_lab
from .lab_to_rgb import lab_to_rgb
from .od_to_rgb import od_to_rgb
from .rgb_to_hsi import rgb_to_hsi
from .rgb_to_lab import rgb_to_lab
# make functions available at the package level using these shadow imports
# since we mostly have one function per file
from .rgb_to_od import rgb_to_od
from .rgb_to_sda import rgb_to_sda
from .sda_to_rgb import sda_to_rgb

# list out things that are available for public use
__all__ = (

    # functions and classes of this package
    'lab_mean_std',
    'lab_to_rgb',
    'od_to_rgb',
    'sda_to_rgb',
    'rgb_to_hsi',
    'rgb_to_lab',
    'rgb_to_od',
    'rgb_to_sda',
)
